'use strict';

var async = require('async');
var fs = require('fs');
var Handlebars = require('handlebars');
var inflection = require('inflection');
var mkdirp = require('mkdirp');
var sysPath = require('path');
var logger = require('loggy');

exports.formatTemplate = function(template, templateData) {
  var compiled, key;
  key = '__TEMPLATE_FORMATTER';
  compiled = Handlebars.compile(template.replace(/\\\{/, key));
  return compiled(templateData).toString().replace(key, '\\');
};

Handlebars.registerHelper('camelize', (function() {
  var camelize;
  camelize = function(string) {
    var regexp, rest;
    regexp = /[-_]([a-z])/g;
    rest = string.replace(regexp, function(match, char) {
      return char.toUpperCase();
    });
    return rest[0].toUpperCase() + rest.slice(1);
  };
  return function(options) {
    return new Handlebars.SafeString(camelize(options.fn(this)));
  };
})());

exports.generateFile = function(path, data, callback) {
  fs.exists(path, function(exists) {
    var parentDir, write;
    if (exists) {
      logger.info("skipping " + path + " (already exists)");
      if (callback != null) {
        return callback();
      }
    } else {
      parentDir = sysPath.dirname(path);
      write = function() {
        logger.info("create " + path);
        fs.writeFile(path, data, callback);
      };
      fs.exists(parentDir, function(exists) {
        if (exists) return write();
        logger.info("init " + parentDir);
        // chmod 755.
        mkdirp(parentDir, 0x1ed, function(error) {
          if (error != null) return logger.error(error);
          write();
        });
      });
    }
  });
};

exports.destroyFile = function(path, callback) {
  fs.unlink(path, function(error) {
    if (error != null) {
      callback(error);
      return logger.error("" + error);
    }
    logger.info("destroy " + path);
    callback();
  });
};

exports.scaffoldFile = function(revert, from, to, templateData, parentPath, callback) {
  if (parentPath) {
    to = sysPath.join(parentPath, sysPath.basename(to));
  }
  if (revert) {
    exports.destroyFile(to, callback);
  } else {
    fs.readFile(from, 'utf8', function(error, contents) {
      var formatted = (function() {
        try {
          return exports.formatTemplate(contents, templateData);
        } catch (error) {
          return contents;
        }
      })();
      exports.generateFile(to, formatted, callback);
    });
  }
};

exports.scaffoldFiles = function(revert, templateData, parentPath) {
  return function(generator, callback) {
    async.forEach(generator.files, function(args, next) {
      exports.scaffoldFile(
        revert, args.from, args.to, templateData, parentPath, next
      );
    }, callback);
  };
};

exports.isDirectory = function(generatorsPath) {
  return function(path, callback) {
    fs.stat(sysPath.join(generatorsPath, path), function(error, stats) {
      if (error != null) logger.error(error);
      callback(stats.isDirectory());
    });
  };
};

exports.readGeneratorConfig = function(generatorsPath) {
  return function(name, callback) {
    var json, path;
    var path = sysPath.resolve(sysPath.join(generatorsPath, name, 'generator.json'));
    var json = require(path);
    json.name = name;
    callback(null, json);
  };
};

exports.formatGeneratorConfig = function(path, json, templateData) {
  var join = function(file) {
    return sysPath.join(path, file);
  };
  var replaceSlashes = function(string) {
    if (sysPath.sep === '\\') {
      return string.replace(/\//g, '\\');
    } else {
      return string;
    }
  };

  json.files = json.files.map(function(object) {
    return {
      from: join(replaceSlashes(object.from)),
      to: replaceSlashes(exports.formatTemplate(object.to, templateData))
    };
  });

  json.dependencies = json.dependencies.map(function(object) {
    return {
      name: object.name,
      params: exports.formatTemplate(object.params, templateData)
    };
  });

  return Object.freeze(json);
};

exports.getDependencyTree = function(generators, generatorName, memo) {
  if (memo == null) memo = [];
  var generator = generators.filter(function(gen) {
    return gen.name === generatorName;
  })[0];
  if (generator == null) {
    throw new Error("Invalid generator " + generatorName);
  }
  (generator.dependencies || []).forEach(function(dependency) {
    exports.getDependencyTree(generators, dependency.name, memo);
  });
  memo.push(generator);
  return memo;
};

exports.generateFiles = function(revert, generatorsPath, type, templateData, parentPath, callback) {
  fs.readdir(generatorsPath, function(error, files) {
    if (error != null) throw new Error(error);

    // Get directories from generators directory.
    async.filter(files, exports.isDirectory(generatorsPath), function(directories) {

      // Read all generator configs.
      async.map(directories, exports.readGeneratorConfig(generatorsPath), function(error, configs) {
        if (error != null) throw new Error(error);
        var generators = directories.map(function(directory, index) {
          var path;
          path = sysPath.join(generatorsPath, directory);
          return exports.formatGeneratorConfig(path, configs[index], templateData);
        });

        // Calculate dependency trees, do the scaffolding.
        var tree = exports.getDependencyTree(generators, type);
        async.forEach(tree, exports.scaffoldFiles(revert, templateData, parentPath), function(error) {
          if (error != null) return callback(error);
          callback();
        });
      });
    });
  });
};

var scaffolt = module.exports = function(type, name, options, callback) {
  // Set some default params.
  if (options == null) options = {};
  if (callback == null) callback = function() {};

  var pluralName = options.pluralName;
  var generatorsPath = options.generatorsPath;
  var revert = options.revert;
  var parentPath = options.parentPath;

  if (pluralName == null) pluralName = inflection.pluralize(name);
  if (generatorsPath == null) generatorsPath = 'generators';
  if (revert == null) revert = false;
  var templateData = {name: name, pluralName: pluralName};

  exports.generateFiles(revert, generatorsPath, type, templateData, parentPath, function(error) {
    if (error != null) {
      logger.error(error);
      return callback(error);
    }
    callback();
  });
};
